from __future__ import annotations

import numpy as np

from . import text
from . import overlay


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

class BarOverlay(overlay.Overlay):
    """Overlay that implements a bar."""

    def __init__(self, *args, x0=0.1, y0=0.1, height_pixels=20, color=(1, 1, 1, 1), initial_length=0.2, **kwargs):
        self.x0 = x0
        self.y0 = y0
        self.height_pixels = height_pixels
        self.color = color
        self.length = initial_length # update just by directly setting self.length in clipspace coords - simple :-)

        super().__init__(*args, **kwargs)

    def render_contents(self) -> np.ndarray:
        # just a single pixel of the right color
        pixel = np.ones((1, 1, 4), dtype=np.float32)
        pixel[0,0,:] = self.color
        return pixel

    def get_clipspace_coordinates(self, window_pixel_width, window_pixel_height):
        height_clipspace = 2.0 * self.height_pixels / window_pixel_height
        return self.x0, self.y0, self.length, height_clipspace


class ScalebarOverlay:
    def __init__(self, visualizer: Visualizer):
        self._label = text.TextOverlay(visualizer, "Scalebar", (-0.9, -0.85), 80, color=(1, 1, 1, 1))
        self._bar = BarOverlay(visualizer, x0=-0.9, y0=-0.9, height_pixels=10, color=(1, 1, 1, 1))
        self._visualizer = visualizer

    def encode_render_pass(self, command_encoder: wgpu.GPUCommandEncoder, target_texture_view: wgpu.GPUTextureView):
        physical_scalebar_length = self._recommend_physical_scalebar_length()
        self._bar.length = physical_scalebar_length / self._visualizer.scale
        # note that the visualizer scale refers to a square rendering target
        # however only part of this is shown in the final window if the window
        # aspect ratio isn't 1:1. So we now need to correct for this effect.
        # The full x extent is shown if the width is greater than the height, so
        # no correction is needed then. If the height is greater than the width,
        # then the x extent is scaled by the ratio of the height to the width.

        if self._visualizer.canvas.width_physical < self._visualizer.canvas.height_physical:
            self._bar.length *= self._visualizer.canvas.height_physical / self._visualizer.canvas.width_physical

        self._update_scalebar_label(physical_scalebar_length)

        self._label.encode_render_pass(command_encoder, target_texture_view)
        self._bar.encode_render_pass(command_encoder, target_texture_view)
    def _get_scalebar_label_text(self, physical_scalebar_length_kpc):
        if physical_scalebar_length_kpc < 1:
            return f"{physical_scalebar_length_kpc * 1000:.0f} pc"
        if physical_scalebar_length_kpc < 1000:
            return f"{physical_scalebar_length_kpc:.0f} kpc"
        else:
            return f"{physical_scalebar_length_kpc / 1000:.0f} Mpc"

    def _update_scalebar_label(self, physical_scalebar_length):
        if getattr(self, "_scalebar_label_is_for_length", None) != physical_scalebar_length:
            self._label.text = self._get_scalebar_label_text(physical_scalebar_length)
            self._scalebar_label_is_for_length = physical_scalebar_length
            self._label.update()

    def _recommend_physical_scalebar_length(self):
        # target is for the scalebar to be no more than 1/2 the viewport
        # (but not too much less either); however the length is to be 10^n or 5*10^n
        # in world coordinates the viewport is 2 * self.scale kpc wide
        # so the maximum scalebar length should be self.scale kpc

        physical_scalebar_length =  self._visualizer.scale
        # now quantize it:
        power_of_ten = np.floor(np.log10(physical_scalebar_length))
        mantissa = physical_scalebar_length / 10 ** power_of_ten
        if mantissa < 2.0:
            physical_scalebar_length = 10.0 ** power_of_ten
        elif mantissa < 5.0:
            physical_scalebar_length = 2.0 * 10.0 ** power_of_ten
        else:
            physical_scalebar_length = 5.0 * 10.0 ** power_of_ten
        return physical_scalebar_length




from __future__ import annotations

import logging
import numpy as np
import time
import wgpu

from contextlib import contextmanager

from . import config
from . import canvas
from . import colormap
from . import multiresolution_sph, sph, periodic_sph
from . import colorbar
from . import text
from . import scalebar
from . import loader
from . import util
from . import line
from . import simcube
from . import view_synchronizer
from .drawreason import DrawReason

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
class VisualizerBase:
    colorbar_aspect_ratio = config.COLORBAR_ASPECT_RATIO
    show_status = True
    device = None # device will be shared across all instances

    def __init__(self, data_loader_class = loader.TestDataLoader, data_loader_args = (),
                 *, render_resolution = config.DEFAULT_RESOLUTION, periodic_tiling = False,
                 colormap_name = config.DEFAULT_COLORMAP, canvas_class = canvas.VisualizerCanvas):
        self._colormap_name = colormap_name
        self._render_resolution = render_resolution
        self.crosshairs_visible = False

        self._prevent_sph_rendering = False # when True, prevents the sph from rendering, to ensure quick screen updates
        self.vmin_vmax_is_set = False

        self.show_colorbar = True
        self.show_scalebar = True

        self.canvas = canvas_class(visualizer=self, title="topsy")

        self._setup_wgpu()

        self.data_loader = data_loader_class(self.device, *data_loader_args)

        self.periodicity_scale = self.data_loader.get_periodicity_scale()

        self._colormap = colormap.Colormap(self, weighted_average = False)
        self._periodic_tiling = periodic_tiling

        if periodic_tiling:
            self._sph = periodic_sph.PeriodicSPH(self, self.render_texture)
        else:
            self._sph = sph.SPH(self, self.render_texture)
        #self._sph = multiresolution_sph.MultiresolutionSPH(self, self.render_texture)

        self._last_status_update = 0.0
        self._status = text.TextOverlay(self, "topsy", (-0.9, 0.9), 80, color=(1, 1, 1, 1))

        self._colorbar = colorbar.ColorbarOverlay(self, 0.0, 1.0, self.colormap_name, "TODO")
        self._scalebar = scalebar.ScalebarOverlay(self)

        self._crosshairs = line.Line(self,
                                     [(-1, 0,0,0), (1, 0,0,0),
                                      (200,200,0,0),
                                      (0, 1, 0, 0), (0, -1, 0, 0)],
                                     (1, 1, 1, 0.3) # color
                                     , 10.0)
        self._cube = simcube.SimCube(self, (1, 1, 1, 0.3), 10.0)

        self._render_timer = util.TimeGpuOperation(self.device)

        self.invalidate(DrawReason.INITIAL_UPDATE)

    def _setup_wgpu(self):
        self.adapter: wgpu.GPUAdapter = wgpu.gpu.request_adapter(power_preference="high-performance")
        if self.device is None:
            type(self).device: wgpu.GPUDevice = self.adapter.request_device(
                required_features=["TextureAdapterSpecificFormatFeatures", "float32-filterable"])
        self.context: wgpu.GPUCanvasContext = self.canvas.get_context()
        self.canvas_format = self.context.get_preferred_format(self.adapter)
        if self.canvas_format.endswith("-srgb"):
            # matplotlib colours aren't srgb. It might be better to convert
            # but for now, just stop the canvas being srgb
            self.canvas_format = self.canvas_format[:-5]
        self.context.configure(device=self.device, format=self.canvas_format)
        self.render_texture: wgpu.GPUTexture = self.device.create_texture(
            size=(self._render_resolution, self._render_resolution, 1),
            usage=wgpu.TextureUsage.RENDER_ATTACHMENT |
                  wgpu.TextureUsage.TEXTURE_BINDING |
                  wgpu.TextureUsage.COPY_SRC,
            format=wgpu.TextureFormat.rg32float,
            label="sph_render_texture",
        )

    def invalidate(self, reason=DrawReason.CHANGE):
        # NB no need to check if we're already pending a draw - wgpu.gui does that for us
        self.canvas.request_draw(lambda: self.draw(reason))

    def rotate(self, x_angle, y_angle):
        dx_rotation_matrix = self._x_rotation_matrix(x_angle)
        dy_rotation_matrix = self._y_rotation_matrix(y_angle)
        self.rotation_matrix = dx_rotation_matrix @ dy_rotation_matrix @ self.rotation_matrix

    @property
    def rotation_matrix(self):
        return self._sph.rotation_matrix

    @rotation_matrix.setter
    def rotation_matrix(self, value):
        self._sph.rotation_matrix = value
        self.invalidate()

    @property
    def colormap_name(self):
        return self._colormap_name

    @colormap_name.setter
    def colormap_name(self, value):
        self._colormap_name = value
        self._reinitialize_colormap_and_bar()
        self.invalidate(reason=DrawReason.PRESENTATION_CHANGE)

    @property
    def position_offset(self):
        return self._sph.position_offset

    @position_offset.setter
    def position_offset(self, value):
        self._sph.position_offset = value
        self.invalidate()

    def reset_view(self):
        self._sph.rotation_matrix = np.eye(3)
        self.scale = config.DEFAULT_SCALE
        self._sph.position_offset = np.zeros(3)

    @property
    def scale(self):
        """Return the scalefactor from kpc to viewport coordinates. Viewport will therefore be 2*scale wide."""
        return self._sph.scale
    @scale.setter
    def scale(self, value):
        self._sph.scale = value
        self.invalidate()

    @property
    def quantity_name(self):
        """The name of the quantity being visualised, or None if density projection."""
        return self.data_loader.quantity_name

    @property
    def averaging(self):
        """True if the quantity being visualised is a weighted average, False if it is a mass projection."""
        return self.data_loader.quantity_name is not None

    @quantity_name.setter
    def quantity_name(self, value):

        if value is not None:
            # see if we can get it. Assume it'll be cached, so this won't waste time.
            try:
                self.data_loader.get_named_quantity(value)
            except Exception as e:
                raise ValueError(f"Unable to get quantity named '{value}'") from e

        self.data_loader.quantity_name = value
        self.vmin_vmax_is_set = False
        self._reinitialize_colormap_and_bar()
        self.invalidate()

    def _reinitialize_colormap_and_bar(self):
        vmin, vmax, log_scale = self.vmin, self.vmax, self.log_scale
        self._colormap = colormap.Colormap(self, weighted_average=self.quantity_name is not None)
        if self.vmin_vmax_is_set:
            self._colormap.vmin = vmin
            self._colormap.vmax = vmax
            self._colormap.log_scale = log_scale
        self._colorbar = colorbar.ColorbarOverlay(self, self.vmin, self.vmax, self.colormap_name,
                                                  self._get_colorbar_label())

    def _get_colorbar_label(self):
        label = self.data_loader.get_quantity_label()
        if self._colormap.log_scale:
            label = r"$\log_{10}$ " + label
        return label

    @staticmethod
    def _y_rotation_matrix(angle):
        return np.array([[1, 0, 0],
                         [0, np.cos(angle), -np.sin(angle)],
                         [0, np.sin(angle), np.cos(angle)]])

    @staticmethod
    def _x_rotation_matrix(angle):
        return np.array([[np.cos(angle), 0, np.sin(angle)],
                         [0, 1, 0],
                         [-np.sin(angle), 0, np.cos(angle)]])

    def _check_whether_inactive(self):
        if time.time()-self._last_lores_draw_time>config.FULL_RESOLUTION_RENDER_AFTER*0.95:
            self._last_lores_draw_time = np.inf # prevent this from being called again
            self.invalidate(reason=DrawReason.REFINE)

    @contextmanager
    def prevent_sph_rendering(self):
        self._prevent_sph_rendering = True
        try:
            yield
        finally:
            self._prevent_sph_rendering = False

    def draw(self, reason, target_texture_view=None):


        if reason == DrawReason.REFINE or reason == DrawReason.EXPORT:
            self._sph.downsample_factor = 1

        if reason!=DrawReason.PRESENTATION_CHANGE and (not self._prevent_sph_rendering):
            ce_label = "sph_render"
            # labelling this is useful for understanding performance in macos instruments
            if self._sph.downsample_factor>1:
                ce_label += f"_ds{self._sph.downsample_factor:d}"
            else:
                ce_label += "_fullres"

            command_encoder : wgpu.GPUCommandEncoder = self.device.create_command_encoder(label=ce_label)
            self._sph.encode_render_pass(command_encoder)

            with self._render_timer:
                self.device.queue.submit([command_encoder.finish()])

        if not self.vmin_vmax_is_set:
            logger.info("Setting vmin/vmax")
            self._colormap.autorange_vmin_vmax()
            self.vmin_vmax_is_set = True
            self._refresh_colorbar()

        command_encoder = self.device.create_command_encoder()

        if target_texture_view is None:
            target_texture_view = self.canvas.get_context().get_current_texture().create_view()


        self._colormap.encode_render_pass(command_encoder, target_texture_view)
        if self.show_colorbar:
            self._colorbar.encode_render_pass(command_encoder, target_texture_view)
        if self.show_scalebar:
            self._scalebar.encode_render_pass(command_encoder, target_texture_view)
        if self.crosshairs_visible:
            self._crosshairs.encode_render_pass(command_encoder, target_texture_view)
        if self._periodic_tiling:
            self._cube.encode_render_pass(command_encoder, target_texture_view)

        if reason == DrawReason.REFINE:
            self.display_status("Full-res render took {:.2f} s".format(self._render_timer.last_duration, timeout=0.1))

        if self.show_status:
            self._update_and_display_status(command_encoder, target_texture_view)

        self.device.queue.submit([command_encoder.finish()])



        if reason != DrawReason.PRESENTATION_CHANGE and reason != DrawReason.EXPORT and (not self._prevent_sph_rendering):
            if self._sph.downsample_factor>1:
                self._last_lores_draw_time = time.time()
                self.canvas.call_later(config.FULL_RESOLUTION_RENDER_AFTER, self._check_whether_inactive)
            elif self._render_timer.last_duration>1/config.TARGET_FPS and self._sph.downsample_factor==1:
                # this will affect the NEXT frame, not this one!
                self._sph.downsample_factor = int(np.floor(float(config.TARGET_FPS)*self._render_timer.last_duration))

    @property
    def vmin(self):
        return self._colormap.vmin

    @property
    def vmax(self):
        return self._colormap.vmax

    @vmin.setter
    def vmin(self, value):
        self._colormap.vmin = value
        self.vmin_vmax_is_set = True
        self._refresh_colorbar()
        self.invalidate()

    @vmax.setter
    def vmax(self, value):
        self._colormap.vmax = value
        self.vmin_vmax_is_set = True
        self._refresh_colorbar()
        self.invalidate()

    @property
    def log_scale(self):
        return self._colormap.log_scale

    @log_scale.setter
    def log_scale(self, value):
        self._colormap.log_scale = value
        self._refresh_colorbar()
        self.invalidate()

    def _refresh_colorbar(self):
        self._colorbar.vmin = self._colormap.vmin
        self._colorbar.vmax = self._colormap.vmax
        self._colorbar.label = self._get_colorbar_label()
        self._colorbar.update()

    def sph_clipspace_to_screen_clipspace_matrix(self):
        aspect_ratio = self.canvas.width_physical / self.canvas.height_physical
        if aspect_ratio>1:
            y_squash = aspect_ratio
            x_squash = 1.0
        elif aspect_ratio<1:
            y_squash = 1.0
            x_squash = 1.0/aspect_ratio
        else:
            x_squash = 1.0
            y_squash = 1.0

        matr = np.eye(4, dtype=np.float32)
        matr[0,0] = x_squash
        matr[1,1] = y_squash
        return matr



    def display_status(self, text, timeout=0.5):
        self._override_status_text = text
        self._override_status_text_until = time.time()+timeout

    def _update_and_display_status(self, command_encoder, target_texture_view):
        now = time.time()
        if hasattr(self, "_override_status_text_until") and now<self._override_status_text_until:
            if self._status.text!=self._override_status_text and now-self._last_status_update>config.STATUS_LINE_UPDATE_INTERVAL_RAPID:
                self._status.text = self._override_status_text
                self._last_status_update = now
                self._status.update()

        elif now - self._last_status_update > config.STATUS_LINE_UPDATE_INTERVAL:
            self._last_status_update = now
            self._status.text = f"${1.0 / self._render_timer.running_mean_duration:.0f}$ fps"
            if self._sph.downsample_factor > 1:
                self._status.text += f", downsample={self._sph.downsample_factor:d}"

            self._status.update()

        self._status.encode_render_pass(command_encoder, target_texture_view)

    def get_sph_image(self) -> np.ndarray:
        im = self.device.queue.read_texture({'texture':self.render_texture, 'origin':(0, 0, 0)},
                                            {'bytes_per_row':8*self._render_resolution},
                                            (self._render_resolution, self._render_resolution, 1))
        np_im = np.frombuffer(im, dtype=np.float32).reshape((self._render_resolution, self._render_resolution, 2))
        if self.averaging:
            im = np_im[:,:,1]/np_im[:,:,0]
        else:
            im = np_im[:,:,0]
        return im

    def get_presentation_image(self) -> np.ndarray:
        texture = self.context.get_current_texture()
        size = texture.size
        bytes_per_pixel = 4 # NB this might be wrong in principle!
        data = self.device.queue.read_texture(
            {
                "texture": texture,
                "mip_level": 0,
                "origin": (0, 0, 0),
            },
            {
                "offset": 0,
                "bytes_per_row": bytes_per_pixel * size[0],
                "rows_per_image": size[1],
            },
            size,
        )

        return np.frombuffer(data, np.uint8).reshape(size[1], size[0], 4)


    def save(self, filename='output.pdf'):
        image = self.get_sph_image()
        import matplotlib.pyplot as p
        fig = p.figure()
        p.clf()
        p.set_cmap(self.colormap_name)
        extent = np.array([-1., 1., -1., 1.])*self.scale
        if self._colormap.log_scale:
            image = np.log10(image)

        p.imshow(image,
                 vmin=self._colormap.vmin,
                 vmax=self._colormap.vmax,
                 extent=extent)
        p.xlabel("$x$/kpc")
        p.colorbar().set_label(self._colorbar.label)
        p.savefig(filename)
        p.close(fig)

    def show(self, force=False):
        from wgpu.gui import jupyter
        if isinstance(self.canvas, jupyter.WgpuCanvas):
            return self.canvas
        else:
            from wgpu.gui import qt # can only safely import this if we think we're running in a qt environment
            assert isinstance(self.canvas, qt.WgpuCanvas)
            self.canvas.show()
            if force or not util.is_inside_ipython():
                qt.run()
            elif not util.is_ipython_running_qt_event_loop():
                # is_inside_ipython_console must be True; if it were False, the previous branch would have run
                # instead.
                print("\r\nYou appear to be running from inside ipython, but the gui event loop is not running.\r\n"
                      "Please run %gui qt in ipython before calling show().\r\n"
                      "\r\n"
                      "Alternatively, if you do not want to continue interacting with ipython while the\r\n"
                      "visualizer is running, you can call show(force=True) to run the gui without access\r\n"
                      "to the ipython console until you close the visualizer window.\r\n\r\n"
                      )
        #else:
        #    raise RuntimeError("The wgpu library is using a gui backend that topsy does not recognize")


class Visualizer(view_synchronizer.SynchronizationMixin, VisualizerBase):
    pass


import matplotlib.pyplot as plt
import matplotlib
import matplotlib.figure as figure
import matplotlib.colors as colors
import numpy as np

from .overlay import Overlay


class ColorbarOverlay(Overlay):
    def __init__(self, visualizer, vmin, vmax, colormap, label, *, dpi_logical=72, **kwargs):
        self.dpi_logical = dpi_logical
        self.kwargs = kwargs
        self._aspect_ratio = 0.2
        self.vmin = vmin
        self.vmax = vmax
        self.colormap = colormap
        self.label = label
        self._last_width = None
        self._last_height = None

        super().__init__(visualizer)

    def get_clipspace_coordinates(self, pixel_width, pixel_height):
        im = self.get_contents()
        height = 2.0
        width = 2.0*pixel_height*im.shape[1]/im.shape[0]/pixel_width
        x,y = 1.0-width,-1.0
        if self._last_width!=pixel_width or self._last_height!=pixel_height:
            # contents is the wrong size
            self.update()
        self._last_width = pixel_width
        self._last_height = pixel_height
        return x, y, width, height
    def render_contents(self):
        dpi_physical = self.dpi_logical*self._visualizer.canvas.pixel_ratio

        fig = figure.Figure(figsize=(self._visualizer.canvas.height_physical * self._aspect_ratio/dpi_physical,
                                  self._visualizer.canvas.height_physical/dpi_physical),
                         dpi=dpi_physical,
                         facecolor=(1.0, 1.0, 1.0, 0.5))

        canvas = matplotlib.backends.backend_agg.FigureCanvasAgg(fig)

        cmap = matplotlib.colormaps[self.colormap]
        cNorm = colors.Normalize(vmin=self.vmin, vmax=self.vmax)
        cb1 = matplotlib.colorbar.ColorbarBase(fig.add_axes([0.05, 0.05, 0.3, 0.9]),
                                               cmap=cmap, norm=cNorm, orientation='vertical')
        cb1.set_label(self.label)

        fig.canvas.draw()
        width,height = fig.canvas.get_width_height(physical=True)

        result: np.ndarray = np.frombuffer(fig.canvas.buffer_rgba(),dtype=np.uint8).reshape((height,width,4)).transpose((1,0,2))
        result = result.swapaxes(0,1).astype(np.float32)/256

        return result



import numpy as np
import time
import wgpu
import re
import os

def load_shader(name):
    from importlib import resources
    with open(resources.files("topsy.shaders") / name, "r") as f:
        return f.read()

def preprocess_shader(shader_code, active_flags):
    """A hacky preprocessor for WGSL shaders.

    Any line in shader_code containing [[FLAG]] will be removed if FLAG is not in active_flags.
    Otherwise, the string [[FLAG]] will be removed, leaving just valid syntax.

    This is needed because we can't use
    const values in the shader yet, so we need to use something like #ifdefs instead.
    In final version of webgpu doesn't look like this will be needed"""
    for f in active_flags:
        shader_code = re.sub(f"^.*\[\[{f}]](.*)$", r"\1", shader_code, flags=re.MULTILINE)
    shader_code = re.sub(r"^.*\[\[[A-Z_]+]].*$", "", shader_code, flags=re.MULTILINE)
    return shader_code

def is_inside_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def is_inside_jupyter_notebook():
    return "JPY_SESSION_NAME" in os.environ

def is_ipython_running_qt_event_loop():
    if not is_inside_ipython():
        return False
    import IPython.lib.guisupport
    return IPython.lib.guisupport.is_event_loop_running_qt4()

def determine_backend():
    if is_inside_ipython():
        pass
class TimeGpuOperation:
    """Context manager for timing GPU operations"""
    def __init__(self, device, n_frames_smooth=10):
        self.device = device
        self.n_frames_smooth = n_frames_smooth
        self._recent_times = []
    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        # Now, we want to measure how much time the render has taken so that we can adapt
        # for the next frame if needed. However, the GPU is asynchronous. In the long term
        # there should be facilities like callbacks or querysets to help with this, but
        # right now these don't seem to be implemented. So we need to make something block
        # until the current queue is complete. The hack here is to do a trivial read
        # operation
        dummy_buffer = self.device.create_buffer(
            size=16,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST | wgpu.BufferUsage.COPY_SRC
        )
        self.device.queue.read_buffer(dummy_buffer, 0)
        end = time.time()

        self.end = time.time()
        self.last_duration = self.end - self.start
        self._recent_times.append(self.last_duration)
        if len(self._recent_times) > self.n_frames_smooth:
            self._recent_times.pop(0)
    @property
    def running_mean_duration(self):
        return np.mean(self._recent_times)



from __future__ import annotations

import numpy as np
import wgpu
from . import sph
from . import overlay
from logging import getLogger

logger = getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

class SPHAccumulationOverlay(overlay.Overlay):
    _blending = (
                    wgpu.BlendFactor.one,
                    wgpu.BlendFactor.one,
                    wgpu.BlendOperation.add,
                 )
    def __init__(self, visualizer: Visualizer, source_texture: wgpu.GPUTexture):
        self._texture = source_texture
        self.num_repetitions = 0
        self.panel_scale = 1.0
        super().__init__(visualizer, source_texture.format)


    def _setup_texture(self):
        pass

    def get_clipspace_coordinates(self, width, height) -> tuple[float, float, float, float]:
        return -1.0, -1.0, 2.0, 2.0

    def get_instance_offsets_and_weights(self):
        offsets = []
        for xoff in range(-self.num_repetitions, self.num_repetitions + 1):
            for yoff in range(-self.num_repetitions, self.num_repetitions + 1):
                offsets.append([xoff*self.panel_scale,yoff*self.panel_scale])

        return np.array(offsets, dtype=np.float32)
    def render_contents(self) -> np.ndarray:
        # must be implemented, but should never be called because texture is provided externally
        raise RuntimeError("SPHAccumulationOverlay.render_contents() should never be called")

class MultiresolutionSPH:
    """A drop-in replacement for the SPH class, which renders to multiple resolutions and then combines them."""

    def __init__(self, visualizer: Visualizer, render_texture: wgpu.GPUTexture, max_pixels=40.0):
        self._downsample_factor = 1
        self._resolution_final = render_texture.width
        assert render_texture.width == render_texture.height
        self._pixel_scaling_factors = [1, 4, 16]
        self._resolutions = [self._resolution_final // factor for factor in self._pixel_scaling_factors]
        self._textures = [render_texture]
        for i,r in enumerate(self._resolutions[1:],1):
            self._textures.append(visualizer.device.create_texture(
                size=(r, r, 1),
                format=self._textures[0].format,
                usage=self._textures[0].usage,
                label=f"multires_sph_level_{i}"
            ))

        self._renderers: list[sph.SPH] = [sph.SPH(visualizer, texture) for texture in self._textures]


        self._accumulators = [SPHAccumulationOverlay(visualizer, self._textures[i])
                              for i in range(len(self._textures)-1,0,-1)]

        layer_counts = {}
        for ds in self._pixel_scaling_factors:
            layer_counts[ds] = layer_counts.get(ds, 0) + 1

        logger.info(f"Multi-resolution SPH initialized: {len(self._renderers)} layers, with pixel downsample factors as follows:")

        current_offsets = {ds: 0 for ds in layer_counts.keys()}

        if self._pixel_scaling_factors[0]!=1:
            raise RuntimeError("Factor 1 must be first entry in the pixel scaling factors list")

        self._original_downsamp_factors = []

        for i in range(len(self._renderers)):
            r = self._renderers[i]
            downsamp_fac = self._pixel_scaling_factors[i]
            if downsamp_fac==1:
                r.min_pixels = 0.0
            else:
                next_highest_pixelscale_fac = max([ds for ds in self._pixel_scaling_factors if ds < downsamp_fac])
                r.min_pixels = max_pixels * next_highest_pixelscale_fac / self._pixel_scaling_factors[i]

            if downsamp_fac==max(self._pixel_scaling_factors):
                r.max_pixels = np.inf
            else:
                r.max_pixels = max_pixels

            r.downsample_offset = current_offsets[self._pixel_scaling_factors[i]]
            current_offsets[self._pixel_scaling_factors[i]] += 1
            r.downsample_factor = layer_counts[self._pixel_scaling_factors[i]]
            self._original_downsamp_factors.append(r.downsample_factor) # so it can be scaled later
            r.mass_scale = 1.0
            logger.info(f"  Layer {i}: pixel_width={self._resolutions[i]} min_pixels={r.min_pixels}, max_pixels={r.max_pixels}, downsample_factor={r.downsample_factor}, downsample_offset={r.downsample_offset}")

        self._visualizer = visualizer

    @property
    def scale(self):
        return self._renderers[0].scale

    @scale.setter
    def scale(self, value):
        for s in self._renderers:
            s.scale = value

    @property
    def downsample_factor(self):
        return self._downsample_factor

    @downsample_factor.setter
    def downsample_factor(self, value):
        self._downsample_factor = value
        for renderer, orig_fac in zip(self._renderers, self._original_downsamp_factors):
            renderer.downsample_factor = value * orig_fac
            renderer.mass_scale = value

    @property
    def rotation_matrix(self):
        return self._renderers[0].rotation_matrix

    @rotation_matrix.setter
    def rotation_matrix(self, value):
        for s in self._renderers:
            s.rotation_matrix = value

    def encode_render_pass(self, command_encoder: wgpu.GPUCommandEncoder):
        for s in self._renderers:
            s.encode_render_pass(command_encoder)

        for a in self._accumulators:
            a.encode_render_pass(command_encoder, self._textures[0].create_view())





from __future__ import annotations

import numpy as np
from wgpu.gui.offscreen import WgpuManualOffscreenCanvas, call_later

from . import VisualizerCanvasBase

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..visualizer import Visualizer


class VisualizerCanvas(VisualizerCanvasBase, WgpuManualOffscreenCanvas):

    @classmethod
    def call_later(cls, delay, fn, *args):
        call_later(delay, fn, *args)




import numpy as np
import wgpu
import logging
import pynbody
import pickle

from typing import Optional

from . import config
# ABC support:
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)
class AbstractDataLoader(ABC):
    def __init__(self, device: wgpu.GPUDevice):
        self._device = device
        self.quantity_name = None
        self._quantity_buffer_is_for_name = None
    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def get_positions(self):
        pass

    @abstractmethod
    def get_smooth(self):
        pass

    @abstractmethod
    def get_mass(self):
        pass

    @abstractmethod
    def get_named_quantity(self, name):
        pass

    @abstractmethod
    def get_quantity_label(self):
        pass

    def get_pos_smooth(self):
        pos_smooth = np.empty((len(self), 4), dtype=np.float32)
        pos_smooth[:, :3] = self.get_positions()
        pos_smooth[:, 3] = self.get_smooth()
        return pos_smooth

    def get_pos_smooth_buffer(self):
        if not hasattr(self, "_pos_smooth_buffer"):
            logger.info("Creating position+smoothing buffer")
            data = self.get_pos_smooth()
            self._pos_smooth_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
        return self._pos_smooth_buffer

    def get_mass_buffer(self):
        if not hasattr(self, "_mass_buffer"):
            logger.info("Creating mass buffer")
            data = self.get_mass()
            self._mass_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
        return self._mass_buffer

    def get_quantity_buffer(self):
        if self.quantity_name is None:
            return self.get_mass_buffer()
        elif self._quantity_buffer_is_for_name != self.quantity_name:
            logger.info(f"Creating {self.quantity_name} buffer")
            data = self.get_named_quantity(self.quantity_name)
            self._named_quantity_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
            self._quantity_buffer_is_for_name = self.quantity_name
        return self._named_quantity_buffer

    def get_periodicity_scale(self):
        return np.inf


class PynbodyDataInMemory(AbstractDataLoader):
    """Base class for data loaders that use pynbody."""

    def __init__(self, device: wgpu.GPUDevice, snapshot: pynbody.snapshot.SimSnap):
        super().__init__(device)

        self.snapshot = snapshot

        # randomize order to avoid artifacts when downsampling number of particles on display
        self._random_order = np.random.permutation(len(self.snapshot))

    def get_positions(self):
        return self.snapshot['pos'].astype(np.float32)[self._random_order]

    def get_smooth(self):
        return self.snapshot['smooth'].astype(np.float32)[self._random_order]

    def get_mass(self):
        return self.snapshot['mass'].astype(np.float32)[self._random_order]

    def get_named_quantity(self, name):
        qty =self.snapshot[name]
        if len(qty.shape)==2:
            qty = qty[:,0]
        return qty.astype(np.float32)[self._random_order]

    def get_quantity_names(self):
        return self.snapshot.loadable_keys()

    def get_quantity_label(self):
        if self.quantity_name is None:
            return r"density / $M_{\odot} / \mathrm{kpc}^2$"
        else:
            lunit = self.snapshot[self.quantity_name].units.latex()
            if lunit != "":
                lunit = "$/" + lunit + "$"
            return self.quantity_name + lunit

    def __len__(self):
        return len(self.snapshot)

    def get_periodicity_scale(self):
        return float(self.snapshot.properties['boxsize'].in_units("kpc"))
    def get_filename(self):
        return self.snapshot.filename

class PynbodyDataLoader(PynbodyDataInMemory):
    """Literal data loader for pynbody (starts from just a filename)"""
    def __init__(self, device: wgpu.GPUDevice, filename: str, center: str, particle: str,
                 take_region: Optional[pynbody.filt.Filter] = None):

        logger.info(f"Data filename = {filename}, center = {center}, particle = {particle}")
        if take_region is None:
            snapshot = pynbody.load(filename)
        else:
            snapshot = pynbody.load(filename, take_region=take_region)

        snapshot.physical_units()
        self.filename = filename

        snapshot = snapshot[pynbody.family.get_family(particle)]

        super().__init__(device, snapshot)

        self._perform_centering(center)
        snapshot.wrap()
        self._perform_smoothing()

    def _perform_centering(self, center):
        logger.info("Performing centering...")
        if center.startswith("halo-"):
            halo_number = int(center[5:])
            h = self.snapshot.ancestor.halos()
            pynbody.analysis.halo.center(h[halo_number], vel=False)

        elif center == 'zoom':
            f_dm = self.snapshot.ancestor.dm
            pynbody.analysis.halo.center(f_dm[f_dm['mass'] < 1.01 * f_dm['mass'].min()])
        elif center == 'all':
            pynbody.analysis.halo.center(self.snapshot)
        elif center == 'none':
            pass
        else:
            raise ValueError("Unknown centering type")

    def _perform_smoothing(self):
        try:
            logger.info("Looking for cached smoothing/density data...")
            smooth = pickle.load(open(self.filename+'-topsy-smooth.pkl', 'rb'))
            if len(smooth) == len(self.snapshot):
                self.snapshot['smooth'] = smooth
            else:
                raise ValueError("Incorrect number of particles in cached smoothing data")
            logger.info("...success!")

            rho = pickle.load(open(self.filename+'-topsy-rho.pkl', 'rb'))
            if len(rho) == len(self.snapshot):
                self.snapshot['rho'] = rho
            else:
                raise ValueError("Incorrect number of particles in cached density data")
        except:
            logger.info("Generating smoothing/density data - this can take a while but will be cached for future runs")
            pickle.dump(self.snapshot['smooth'], open(self.filename+'-topsy-smooth.pkl', 'wb'))
            pickle.dump(self.snapshot['rho'], open(self.filename+'-topsy-rho.pkl', 'wb'))



class TestDataLoader(AbstractDataLoader):
    def __init__(self, device: wgpu.GPUDevice, n_particles: int = config.TEST_DATA_NUM_PARTICLES_DEFAULT):
        self._n_particles = n_particles
        self._gmm_weights = [0.5, 0.4, 0.1] # should sum to 1
        self._gmm_means = np.array([[0.0, 0.0, 0.0],[0.0, 0.0, 0.0],[6.0,10.0,0.0]])
        self._gmm_std = np.array([[20.0, 20.0, 20.0], [4.0, 0.2, 4.0], [2.0,2.0,3.0]])

        self._gmm_pos = self._generate_samples()
        self._gmm_den = self._evaluate_density(self._gmm_pos)
        super().__init__(device)
    def __len__(self):
        return self._n_particles

    def _evaluate_density(self, pos):
        # simple gaussian mixture model, density of particles per unit volume
        den = np.zeros(len(pos))
        for i in range(len(self._gmm_weights)):
            den += self._gmm_weights[i] * np.exp(-np.sum((pos - self._gmm_means[i])**2/self._gmm_std[i]**2, axis=1)) \
                                                 / ((2 * np.pi)**1.5 * np.prod(self._gmm_std[i]))
        return den*self._n_particles

    def _generate_samples(self):
        # simple gaussian mixture model
        pos = np.empty((self._n_particles, 3), dtype=np.float32)
        if self._n_particles==1:
            pos[0] = self._gmm_means[0]
        else:
            offset = 0
            for i in range(len(self._gmm_weights)):
                cpt_len = int(self._n_particles*self._gmm_weights[i])
                pos[offset:offset+cpt_len] = \
                    np.random.normal(size=(cpt_len, 3), scale=1.0).astype(np.float32) * self._gmm_std[np.newaxis,i,:] + self._gmm_means[i]
                offset += cpt_len
            assert offset == self._n_particles
        return np.random.permutation(pos)

    def get_positions(self):

        return self._gmm_pos

    def get_smooth(self):
        sm = 2.0/self._gmm_den**0.333333
        return sm

    def get_mass(self):
        return np.repeat(np.float32(1e-8), self._n_particles)

    def get_named_quantity(self, name):
        if name=="test-quantity":
            return np.sin(self._gmm_pos[:,0])*np.cos(self._gmm_pos[:,1])*np.cos(self._gmm_pos[:,2])
        else:
            raise KeyError("Unknown quantity name")

    def get_quantity_names(self):
        return ["test-quantity"]

    def get_quantity_label(self):
        if self.quantity_name is None:
            return r"test density / $M_{\odot} / \mathrm{kpc}^2$"
        elif self.quantity_name == "test-quantity":
            return "test quantity"
        else:
            return "unknown"

    def get_filename(self):
        return "test data"



from __future__ import annotations

from .line import Line
import numpy as np

class SimCube(Line):
    def __init__(self, visualizer, color, width):
        size = visualizer.data_loader.get_periodicity_scale()
        line_starts_ends = [[0,0,0], [0,0,1],
                            [0,0,0], [0,1,0],
                            [0,0,0], [1,0,0],
                            [1,1,1], [1,1,0],
                            [1,1,1], [1,0,1],
                            [1,1,1], [0,1,1],
                            [0,1,0], [0,1,1],
                            [0,1,0], [1,1,0],
                            [1,0,1], [1,0,0],
                            [1,0,1], [0,0,1],
                            [1,0,0], [1,1,0],
                            [0,1,1], [0,0,1]
                            ]

        line_starts_ends = np.array(line_starts_ends, dtype=np.float32)
        line_starts_ends -= 0.5

        line_starts_ends *= size

        line_starts_ends = np.concatenate([line_starts_ends, np.ones((line_starts_ends.shape[0], 1))], axis=1)

        self._line_starts = np.ascontiguousarray(line_starts_ends[::2,:], dtype=np.float32)
        self._line_ends = np.ascontiguousarray(line_starts_ends[1::2,:], dtype=np.float32)


        super().__init__(visualizer, None, color, width)

    def encode_render_pass(self, command_encoder: wgpu.GPUCommandEncoder,
                                 target_texture_view: wgpu.GPUTextureView):
        self._params["transform"] = self._visualizer._sph.last_transform_params["transform"] @ self._visualizer.sph_clipspace_to_screen_clipspace_matrix()
        super().encode_render_pass(command_encoder, target_texture_view)





from __future__ import annotations

import numpy as np
import wgpu
import pynbody

from .util import load_shader, preprocess_shader
from . import config

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

class SPH:
    render_format = wgpu.TextureFormat.r32float

    def __init__(self, visualizer: Visualizer, render_texture: wgpu.GPUTexture,
                 wrapping = False):
        self._visualizer = visualizer
        self._render_texture = render_texture
        self._device = visualizer.device
        self._wrapping = wrapping
        self._kernel = None

        self._setup_shader_module()
        self._setup_transform_buffer()
        self._setup_kernel_texture()
        self._setup_render_pipeline()

        self.scale = config.DEFAULT_SCALE
        self.min_pixels = 0.0    # minimum size of softening, in pixels, to qualify for rendering
        self.max_pixels = np.inf # maximum size of softening, in pixels, to qualify for rendering
        self.downsample_factor = 1 # number of particles to increment
        self.downsample_offset = 0  # offset to start skipping particles
        self.rotation_matrix = np.eye(3)
        self.position_offset = np.zeros(3)


    def _setup_shader_module(self):
        wrap_flag = "WRAPPING" if self._wrapping else "NO_WRAPPING"
        code = preprocess_shader(load_shader("sph.wgsl"), [wrap_flag])

        self._shader = self._device.create_shader_module(code=code, label="sph")

    def _setup_transform_buffer(self):
        self._transform_buffer = self._device.create_buffer(
            size=(4*4*4) + 4 + 8 + 8 + 12,  # 4x4 float32 matrix + one float32 scale + float32 min,max size + int32 x 2 + padding
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST
        )

    def _setup_render_pipeline(self):
        self._bind_group_layout = \
            self._device.create_bind_group_layout(
                label="sph_bind_group_layout",
                entries=[
                    {
                        "binding": 0,
                        "visibility": wgpu.ShaderStage.VERTEX,
                        "buffer": {"type": wgpu.BufferBindingType.uniform},
                    },
                    {
                        "binding": 1,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "texture": {"sample_type": wgpu.TextureSampleType.float,
                                    "view_dimension": wgpu.TextureViewDimension.d2},
                    },
                    {
                        "binding": 2,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "sampler": {"type": wgpu.SamplerBindingType.filtering},
                    },
                ]
            )

        self._bind_group = \
            self._device.create_bind_group(
                label="sph_bind_group",
                layout=self._bind_group_layout,
                entries=[
                    {"binding": 0,
                     "resource": {
                         "buffer": self._transform_buffer,
                         "offset": 0,
                         "size": self._transform_buffer.size
                      }
                     },
                    {"binding": 1,
                        "resource": self._kernel_texture.create_view(),
                    },
                    {"binding": 2,
                        "resource": self._kernel_sampler,
                    }
                ]
            )

        self._pipeline_layout = \
            self._device.create_pipeline_layout(
                label="sph_pipeline_layout",
                bind_group_layouts=[self._bind_group_layout]
            )

        self._render_pipeline = \
            self._device.create_render_pipeline(
                layout=self._pipeline_layout,
                label="sph_render_pipeline",
                vertex = {
                    "module": self._shader,
                    "entry_point": "vertex_main",
                    "buffers": [
                        {
                            "array_stride": 16,
                            "step_mode": wgpu.VertexStepMode.instance,
                            "attributes": [
                                {
                                    "format": wgpu.VertexFormat.float32x4,
                                    "offset": 0,
                                    "shader_location": 0,
                                }
                            ]
                        },
                        {
                            "array_stride": 4,
                            "step_mode": wgpu.VertexStepMode.instance,
                            "attributes": [
                                {
                                    "format": wgpu.VertexFormat.float32,
                                    "offset": 0,
                                    "shader_location": 1,
                                }
                            ]
                        },
                        {
                            "array_stride": 4,
                            "step_mode": wgpu.VertexStepMode.instance,
                            "attributes": [
                                {
                                    "format": wgpu.VertexFormat.float32,
                                    "offset": 0,
                                    "shader_location": 2,
                                }
                            ]
                        }

                    ]
                },
                primitive={
                    "topology": wgpu.PrimitiveTopology.triangle_list,
                },
                depth_stencil=None,
                multisample=None,
                fragment={
                    "module": self._shader,
                    "entry_point": "fragment_main",
                    "targets": [
                        {
                            "format": wgpu.TextureFormat.rg32float,
                            "blend": {
                                "color": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.one,
                                    "operation": wgpu.BlendOperation.add,
                                },
                                "alpha": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.one,
                                    "operation": wgpu.BlendOperation.add,
                                },
                              }
                        }
                    ]
                }
            )

    def _get_mass_scale(self):
        return getattr(self, "mass_scale", np.float32(self.downsample_factor))

    def _update_transform_buffer(self):
        model_displace = np.array([[1.0, 0, 0, self.position_offset[0]],
                                   [0, 1.0, 0, self.position_offset[1]],
                                   [0, 0, 1.0, self.position_offset[2]],
                                   [0, 0, 0.0, 1.0]])

        # self._transform is the transformation around the origin (fine for opengl)
        # but in webgpu, the clip space in the z direction is [0,1]
        # so we need a matrix that brings z=0. to z=0.5 and squishes the z direction
        # so that the clipping is the same in all dimensions
        clipcoord_displace = np.array([[1.0, 0, 0, 0.0],
                             [0, 1.0, 0, 0.0],
                             [0, 0, 0.5, 0.5],
                             [0, 0, 0.0, 1.0]])
        transform = np.zeros((4, 4))

        transform[:3,:3] = self.rotation_matrix
        rotation_and_scaling = transform / self.scale

        rotation_and_scaling[3, 3] = 1.0  # w should be unchanged after transform

        scaled_displaced_transform = (clipcoord_displace @ rotation_and_scaling @ model_displace).T
        transform_params_dtype = [("transform", np.float32, (4, 4)),
                                  ("scale_factor", np.float32, (1,)),
                                  ("min_max_size", np.float32, (2,)),
                                  ("downsample_factor", np.uint32, (1,)),
                                  ("downsample_offset", np.uint32, (1,)),
                                  ("mass_scale", np.float32, (1,)),
                                  ("boxsize_by_2_clipspace", np.float32, (1,)),
                                  ("padding", np.int32, (1,))]
        transform_params = np.zeros((), dtype=transform_params_dtype)
        transform_params["transform"] = scaled_displaced_transform
        transform_params["scale_factor"] = 1. / self.scale
        transform_params["mass_scale"] = self._get_mass_scale()
        transform_params["boxsize_by_2_clipspace"] = 0.5 * \
                                                     self._visualizer.periodicity_scale / self.scale

        resolution = self._render_texture.width
        assert resolution == self._render_texture.height

        # min_max_size to be sent in viewport coordinates
        transform_params["min_max_size"] = (2.*self.min_pixels/resolution, 2.*self.max_pixels/resolution)
        transform_params["downsample_factor"] = self.downsample_factor
        transform_params["downsample_offset"] = self.downsample_offset

        self.last_transform_params = transform_params

        self._device.queue.write_buffer(self._transform_buffer, 0, transform_params)

    def encode_render_pass(self, command_encoder):
        self._update_transform_buffer()
        view: wgpu.GPUTextureView = self._render_texture.create_view()
        sph_render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": view,
                    "resolve_target": None,
                    "clear_value": (0.0, 0.0, 0.0, 0.0),
                    "load_op": wgpu.LoadOp.clear,
                    "store_op": wgpu.StoreOp.store,
                }
            ]
        )
        sph_render_pass.set_pipeline(self._render_pipeline)

        num_particles_to_render = len(self._visualizer.data_loader)//self.downsample_factor
        start_particles = num_particles_to_render * self.downsample_offset

        sph_render_pass.set_vertex_buffer(0, self._visualizer.data_loader.get_pos_smooth_buffer(),
                                          offset=start_particles*16)
        sph_render_pass.set_vertex_buffer(1, self._visualizer.data_loader.get_mass_buffer(),
                                          offset=start_particles*4)
        sph_render_pass.set_vertex_buffer(2, self._visualizer.data_loader.get_quantity_buffer(),
                                          offset=start_particles*4)
        sph_render_pass.set_bind_group(0, self._bind_group, [], 0, 99)
        sph_render_pass.draw(6, num_particles_to_render, 0, 0)
        sph_render_pass.end()


    def _get_kernel_at_resolution(self, n_samples):
        if self._kernel is None:
            try:
                self._kernel = pynbody.sph.Kernel2D()
            except AttributeError:
                # pynbody v2:
                self._kernel = pynbody.sph.kernels.Kernel2D()

        # sph kernel is sampled at the centre of the pixels, and the full grid ranges from -2 to 2.
        # thus the left hand most pixel is at -2+2/n_samples, and the right hand most pixel is at 2-2/n_samples.
        pixel_centres = np.linspace(-2+2./n_samples, 2-2./n_samples, n_samples)
        x, y = np.meshgrid(pixel_centres, pixel_centres)
        distance = np.sqrt(x ** 2 + y ** 2)

        # TODO: the below could easily be optimized
        kernel_im = np.array([self._kernel.get_value(d) for d in distance.flatten()]).reshape(n_samples, n_samples)

        # make kernel explicitly mass conserving; naive pixelization makes it not automatically do this.
        # It should be normalized such that the integral over the kernel is 1/h^2. We have h=1 here, and the
        # full width is 4h, so the width of a pixel is dx=4/n_samples. So we need to multiply by dx^2=(n_samples/4)^2.
        # This results in a correction of a few percent, typically; not huge but not negligible either.
        #
        # (Obviously h!=1 generally, so the h^2 normalization occurs within the shader later.)
        kernel_im *= (n_samples / 4) ** 2 / kernel_im.sum()

        return kernel_im

    def _setup_kernel_texture(self, n_samples=64, n_mip_levels = 4):
        if hasattr(SPH, "_kernel_texture"):
            # we only do this once, even if multiple SPH objects (i.e. multi-resolution) is in play
            return

        SPH._kernel_texture = self._device.create_texture(
            label="kernel_texture",
            size=(n_samples, n_samples, 1),
            dimension=wgpu.TextureDimension.d2,
            format=wgpu.TextureFormat.r32float,
            mip_level_count=n_mip_levels,
            sample_count=1,
            usage=wgpu.TextureUsage.COPY_DST | wgpu.TextureUsage.TEXTURE_BINDING,
        )


        for i in range(0, n_mip_levels):
            self._device.queue.write_texture(
                {
                    "texture": self._kernel_texture,
                    "mip_level": i,
                    "origin": (0, 0, 0),
                },
                self._get_kernel_at_resolution(n_samples//2**i).astype(np.float32).tobytes(),
                {
                    "offset": 0,
                    "bytes_per_row": 4 * n_samples // 2**i,
                },
                (n_samples//2**i, n_samples//2**i, 1)
            )


        SPH._kernel_sampler = self._device.create_sampler(label="kernel_sampler",
                                                           mag_filter=wgpu.FilterMode.linear, )


from __future__ import annotations

import logging
import numpy as np
import re
import wgpu
import matplotlib

from . import config

from .util import load_shader, preprocess_shader

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class Colormap:

    def __init__(self, visualizer: Visualizer, weighted_average: bool = False):
        self._visualizer = visualizer
        self._device = visualizer.device
        self._colormap_name = visualizer.colormap_name
        self._input_texture = visualizer.render_texture
        self._output_format = visualizer.canvas_format
        self._weighted_average = weighted_average

        self.vmin, self.vmax = 0,1
        self._log_scale = True
        # all three of these will be reset by set_vmin_vmax

        self._setup_texture()
        self._setup_shader_module()
        self._setup_render_pipeline()

    @property
    def log_scale(self):
        return self._log_scale

    @log_scale.setter
    def log_scale(self, value):
        old_value = self._log_scale
        self._log_scale = value
        if value != old_value:
            self._setup_shader_module()
            self._setup_render_pipeline()

    def _setup_shader_module(self):
        shader_code = load_shader("colormap.wgsl")
        # hack because at present we can't use const values in the shader to compile
        mode = "WEIGHTED_MEAN" if self._weighted_average else "DENSITY"
        active_flags = [mode]

        if self.log_scale:
            active_flags.append("LOG_SCALE")

        shader_code = preprocess_shader(shader_code, active_flags)

        self._shader = self._device.create_shader_module(code=shader_code, label="colormap")



    def _setup_texture(self, num_points=config.COLORMAP_NUM_SAMPLES):
        cmap = matplotlib.colormaps[self._colormap_name]
        rgba = cmap(np.linspace(0.001, 0.999, num_points)).astype(np.float32)

        self._texture = self._device.create_texture(
            label="colormap_texture",
            size=(num_points, 1, 1),
            dimension=wgpu.TextureDimension.d1,
            format=wgpu.TextureFormat.rgba32float,
            mip_level_count=1,
            sample_count=1,
            usage=wgpu.TextureUsage.COPY_DST | wgpu.TextureUsage.TEXTURE_BINDING
        )

        self._device.queue.write_texture(
            {
                "texture": self._texture,
                "mip_level": 0,
                "origin": [0, 0, 0],
            },
            rgba.tobytes(),
            {
                "bytes_per_row": 4 * 4 * num_points,
                "offset": 0,
            },
            (num_points, 1, 1)
        )

    def _setup_render_pipeline(self):
        self._parameter_buffer = self._device.create_buffer(size =4 * 3,
                                                            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST)

        self._bind_group_layout = \
            self._device.create_bind_group_layout(
                label="colormap_bind_group_layout",
                entries=[
                    {
                        "binding": 0,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "texture": {"sample_type": wgpu.TextureSampleType.float,
                                    "view_dimension": wgpu.TextureViewDimension.d2},
                    },
                    {
                        "binding": 1,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "sampler": {"type": wgpu.SamplerBindingType.filtering},
                    },
                    {
                        "binding": 2,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "texture": {"sample_type": wgpu.TextureSampleType.float,
                                    "view_dimension": wgpu.TextureViewDimension.d1},
                    },
                    {
                        "binding": 3,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "sampler": {"type": wgpu.SamplerBindingType.filtering},
                    },
                    {
                        "binding": 4,
                        "visibility": wgpu.ShaderStage.FRAGMENT | wgpu.ShaderStage.VERTEX,
                        "buffer": {"type": wgpu.BufferBindingType.uniform}
                    }
                ]
            )

        self._input_interpolation = self._device.create_sampler(label="colormap_sampler",
                                                                mag_filter=wgpu.FilterMode.linear, )

        self._bind_group = \
            self._device.create_bind_group(
                label="colormap_bind_group",
                layout=self._bind_group_layout,
                entries=[
                    {"binding": 0,
                     "resource": self._input_texture.create_view(),
                     },
                    {"binding": 1,
                     "resource": self._input_interpolation,
                     },
                    {"binding": 2,
                     "resource": self._texture.create_view(),
                     },
                    {"binding": 3,
                     "resource": self._input_interpolation,
                     },
                    {"binding": 4,
                        "resource": {"buffer": self._parameter_buffer,
                                     "offset": 0,
                                     "size": self._parameter_buffer.size}
                     }
                ]
            )

        self._pipeline_layout = \
            self._device.create_pipeline_layout(
                label="colormap_pipeline_layout",
                bind_group_layouts=[self._bind_group_layout]
            )


        self._pipeline = \
            self._device.create_render_pipeline(
                layout=self._pipeline_layout,
                label="colormap_pipeline",
                vertex={
                    "module": self._shader,
                    "entry_point": "vertex_main",
                    "buffers": []
                },
                primitive={
                    "topology": wgpu.PrimitiveTopology.triangle_strip,
                },
                depth_stencil=None,
                multisample=None,
                fragment={
                    "module": self._shader,
                    "entry_point": "fragment_main",
                    "targets": [
                        {
                            "format": self._output_format,
                            "blend": {
                                "color": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.zero,
                                    "operation": wgpu.BlendOperation.add,
                                },
                                "alpha": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.zero,
                                    "operation": wgpu.BlendOperation.add,
                                },
                            }
                        }
                    ]
                }
            )

    def encode_render_pass(self, command_encoder, target_texture_view):
        self._update_parameter_buffer(target_texture_view.size[0], target_texture_view.size[1])
        colormap_render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": target_texture_view,
                    "resolve_target": None,
                    "clear_value": (0.0, 0.0, 0.0, 1.0),
                    "load_op": wgpu.LoadOp.load,
                    "store_op": wgpu.StoreOp.store,
                }
            ]
        )
        colormap_render_pass.set_pipeline(self._pipeline)
        colormap_render_pass.set_bind_group(0, self._bind_group, [], 0, 99)
        colormap_render_pass.draw(4, 1, 0, 0)
        colormap_render_pass.end()


    def autorange_vmin_vmax(self):
        """Set the vmin and vmax values for the colomap based on the most recent SPH render"""

        # This can and probably should be done on-GPU using a compute shader, but for now
        # we'll do it on the CPU
        vals = self._visualizer.get_sph_image().ravel()

        if (vals<0).any():
            self.log_scale = False
        else:
            self.log_scale = True
        # NB above switching of log scale will automatically rebuild the pipeline if needed

        if self.log_scale:
            vals = np.log10(vals)

        vals = vals[np.isfinite(vals)]
        if len(vals) > 200:
            self.vmin, self.vmax = np.percentile(vals, [1.0, 99.9])
        elif len(vals)>2:
            self.vmin, self.vmax = np.min(vals), np.max(vals)
        else:
            logger.warning(
                "Problem setting vmin/vmax, perhaps there are no particles or something is wrong with them?")
            logger.warning("Press 'r' in the window to try again")
            self.vmin, self.vmax = 0.0, 1.0


    def _update_parameter_buffer(self, width, height):
        parameter_dtype = [("vmin", np.float32, (1,)),
                           ("vmax", np.float32, (1,)),
                           ("window_aspect_ratio", np.float32, (1,))]

        parameters = np.zeros((), dtype=parameter_dtype)
        parameters["vmin"] = self.vmin
        parameters["vmax"] = self.vmax


        parameters["window_aspect_ratio"] = float(width)/height
        self._device.queue.write_buffer(self._parameter_buffer, 0, parameters)